use crate::app_interface::Pokemon;
use alloc::rc::Rc;

pub trait PartyTrait {
    fn get_pokemon(&self, index: usize) -> Option<Pokemon>;
    fn length(&self) -> usize;
}

pub type Party = Rc<dyn PartyTrait>;

#[cfg(not(feature = "mock_data"))]
mod implementation {
    use super::*;
    use crate::app_interface::PokemonImpl;
    use crate::handling::extern_ref::{ExternRef, ExternalReferenceType};

    #[derive(Clone)]
    pub struct PartyImpl {
        reference: ExternRef<Self>,
    }

    impl PartyImpl {
        pub fn new(reference: ExternRef<PartyImpl>) -> Self {
            Self { reference }
        }
    }

    impl PartyTrait for PartyImpl {
        fn get_pokemon(&self, index: usize) -> Option<Pokemon> {
            unsafe {
                let v = party_get_pokemon(self.reference, index).get_value();
                if let Some(v) = v {
                    Some(Rc::new(v))
                } else {
                    None
                }
            }
        }

        fn length(&self) -> usize {
            unsafe { party_get_length(self.reference) }
        }
    }

    impl ExternalReferenceType for PartyImpl {
        fn from_extern_value(reference: ExternRef<Self>) -> Self {
            Self::new(reference)
        }
    }

    extern "wasm" {
        fn party_get_pokemon(r: ExternRef<PartyImpl>, index: usize) -> ExternRef<PokemonImpl>;
        fn party_get_length(r: ExternRef<PartyImpl>) -> usize;
    }
}

#[cfg(not(feature = "mock_data"))]
pub use implementation::*;
